package tinyvm

import java.time.Instant
import kotlin.test.Test
import kotlin.test.assertEquals

internal class RepositoryTest {
    private val repository = Repository()
    private val commits = listOf(
        Commit(
            tree = Tree(mapOf("test1.txt" to Blob("Hello World!"))),
            author = Author("Gleb Koval", "gleb@koval.net"),
            message = "Add test1.txt",
            timestamp = Instant.ofEpochSecond(0)
        ), Commit(
            tree = Tree(
                mapOf(
                    "dir1" to Tree(
                        mapOf(
                            "test1.txt" to Blob("Hello World!")
                        )
                    ),
                    "dir2" to Tree(
                        mapOf(
                            "test2.txt" to Blob("This is a second file")
                        )
                    )
                )
            ),
            author = Author("Gleb Koval", "gleb@koval.net"),
            message = "Move test1.txt and add dir2/test2.txt",
            timestamp = Instant.ofEpochSecond(50)
        ), Commit(
            tree = Tree(
                mapOf(
                    "dir1" to Tree(
                        mapOf(
                            "test1.txt" to Blob("Hello World!")
                        )
                    ),
                    "dir2" to Tree(
                        mapOf(
                            "test2.txt" to Blob("This is a second file")
                        )
                    ),
                    "README.md" to Blob("# This is a test repo!")
                )
            ),
            author = Author("Gleb Koval", "gleb@koval.net"),
            message = "Add README.md",
            timestamp = Instant.ofEpochSecond(100)
        )
    )

    @Test
    fun `can commit`() {
        val committed = repository.commit(commits[0])
        assertEquals(commits[0].hash(), committed.hash())
        assertEquals(1, repository.listCommits().size)
        assertEquals(3, repository.dumpObjects().size)
    }

    @Test
    fun `can deduplicate commit`() {
        repository.commit(commits[0])
        assertEquals(1, repository.listCommits().size)
        val committed = repository.commit(commits[0])
        assertEquals(1, repository.listCommits().size)
        assertEquals(3, repository.dumpObjects().size)
        assertEquals(commits[0].hash(), committed.hash())
    }

    @Test
    fun `can commit twice and deduplicate objects`() {
        repository.commit(commits[0])
        val committed = repository.commit(commits[1])
        assertEquals(2, repository.listCommits().size)
        assertEquals(7, repository.dumpObjects().size)
        assertEquals(commits[1].hash(), committed.hash())
        assertEquals(commits.take(2).map { it.hash() }, repository.listCommits().map { it.hash() })
    }

    @Test
    fun `can commit in-between existing commits`() {
        repository.commit(commits[0])
        repository.commit(commits[2])
        val committed = repository.commit(commits[1])
        assertEquals(3, repository.listCommits().size)
        assertEquals(10, repository.dumpObjects().size)
        assertEquals(commits[1].hash(), committed.hash())
        assertEquals(commits.map { it.hash() }, repository.listCommits().map { it.hash() })
    }

    @Test
    fun `can get commit by hash`() {
        repository.commit(commits[0])
        val committed = repository.commit(commits[1])
        repository.commit(commits[2])
        assertEquals(committed, repository.getCommit(commits[1].hash()))
        assertEquals(null, repository.getCommit("00000000000000000000"))
    }

    @Test
    fun `can find commit by predicate`() {
        repository.commit(commits[0])
        val committed = repository.commit(commits[1])
        repository.commit(commits[2])
        assertEquals(committed, repository.findCommit { it.message.matches("Move.*".toRegex()) })
        assertEquals(null, repository.getCommit("00000000000000000000"))
    }
}